#include <libopencm3/stm32/gpio.h>

#include "fan_ctrl_dc.h"
#include "pinout.h"

uint8_t m_last_step;
bool m_enabled;

void fan_ctrl_dc_init(void)
{
	m_last_step = 0;
	m_enabled = false;

	// configure the GPIOs: DC/DC disabled by default
	gpio_clear(DCDC_PORT, DCDC_EN_PIN);
	gpio_mode_setup(DCDC_PORT, GPIO_MODE_OUTPUT, GPIO_PUPD_NONE, DCDC_EN_PIN);

	// configure the feedback resistor GPIOs. These are used as open-drain
	// outputs, so either they are configured as input or as output, but the
	// only valid output value is 0.
	gpio_clear(DCDC_PORT, DCDC_CTRL_FB0_PIN | DCDC_CTRL_FB1_PIN | DCDC_CTRL_FB2_PIN | DCDC_CTRL_FB3_PIN);
	gpio_mode_setup(DCDC_PORT, GPIO_MODE_INPUT, GPIO_PUPD_NONE,
			DCDC_CTRL_FB0_PIN | DCDC_CTRL_FB1_PIN | DCDC_CTRL_FB2_PIN | DCDC_CTRL_FB3_PIN);

	// disable the DC/DC bypass by default (only enabled at 100% PWM)
	gpio_clear(DCDC_PORT, DCDC_BYPASS_SWITCH_PIN);
	gpio_mode_setup(DCDC_PORT, GPIO_MODE_OUTPUT, GPIO_PUPD_NONE, DCDC_BYPASS_SWITCH_PIN);
}


void fan_ctrl_dc_enable(void)
{
	// only set the enable flag. Pins will be set on the next call to fan_ctrl_dc_set_duty().
	m_enabled = true;
	m_last_step = 0;
}


void fan_ctrl_dc_disable(void)
{
	m_enabled = false;

	// shut down the DC/DC converter, disable DC/DC bypass and disable all feedback resistors.
	gpio_clear(DCDC_PORT, DCDC_EN_PIN);
	gpio_clear(DCDC_PORT, DCDC_BYPASS_SWITCH_PIN);
	gpio_mode_setup(DCDC_PORT, GPIO_MODE_INPUT, GPIO_PUPD_NONE,
			DCDC_CTRL_FB0_PIN | DCDC_CTRL_FB1_PIN | DCDC_CTRL_FB2_PIN | DCDC_CTRL_FB3_PIN);
}


void fan_ctrl_dc_set_duty(uint8_t percent)
{
	if(!m_enabled) {
		return;
	}

	// note: as we only have 4 switchable feedback resistors, there are only 16
	// voltage steps available.

	uint8_t step = 15 * (uint16_t)percent / 100;

	if(step == m_last_step) {
		// skip adjustment if the target step has not changed.
		return;
	}

	// disable the DC/DC while the feedback resistors are adjusted
	gpio_clear(DCDC_PORT, DCDC_EN_PIN);
	gpio_clear(DCDC_PORT, DCDC_BYPASS_SWITCH_PIN);

	// if a resistor is active, the output voltage increases. The smallest
	// resistor (FB0) has the strongest influence on the output voltage.
	// Therefore, FB0 corresponds to the MSB of the calculated step.

	if(step & (1 << 3)) {
		gpio_mode_setup(DCDC_PORT, GPIO_MODE_OUTPUT, GPIO_PUPD_NONE, DCDC_CTRL_FB0_PIN);
	} else {
		gpio_mode_setup(DCDC_PORT, GPIO_MODE_INPUT, GPIO_PUPD_NONE, DCDC_CTRL_FB0_PIN);
	}

	if(step & (1 << 2)) {
		gpio_mode_setup(DCDC_PORT, GPIO_MODE_OUTPUT, GPIO_PUPD_NONE, DCDC_CTRL_FB1_PIN);
	} else {
		gpio_mode_setup(DCDC_PORT, GPIO_MODE_INPUT, GPIO_PUPD_NONE, DCDC_CTRL_FB1_PIN);
	}

	if(step & (1 << 1)) {
		gpio_mode_setup(DCDC_PORT, GPIO_MODE_OUTPUT, GPIO_PUPD_NONE, DCDC_CTRL_FB2_PIN);
	} else {
		gpio_mode_setup(DCDC_PORT, GPIO_MODE_INPUT, GPIO_PUPD_NONE, DCDC_CTRL_FB2_PIN);
	}

	if(step & (1 << 0)) {
		gpio_mode_setup(DCDC_PORT, GPIO_MODE_OUTPUT, GPIO_PUPD_NONE, DCDC_CTRL_FB3_PIN);
	} else {
		gpio_mode_setup(DCDC_PORT, GPIO_MODE_INPUT, GPIO_PUPD_NONE, DCDC_CTRL_FB3_PIN);
	}

	// Feedback resistors are now adjusted. Re-enable the DC/DC or, at 100%, set the bypass.
	if(percent >= 100) {
		gpio_set(DCDC_PORT, DCDC_BYPASS_SWITCH_PIN);
	} else {
		gpio_set(DCDC_PORT, DCDC_EN_PIN);
	}
}
